"""
*"The Grammar of Graphics (2005)"* by **Wilkinson**, **Anand** and
**Grossman** describes three types of transformations.

* *Variable transformations* - Used to make statistical operations on
  variables appropriate and meaningful. They are also used to new
  variables.
* *Scale transformations* - Used to make statistical objects displayed
  on dimensions appropriate and meaningful.
* *Coordinate transformations* - Used to manipulate the geometry of
  graphics to help perceive relationships and find meaningful structures
  for representing variations.

`Variable` and `scale` transformations are similar in-that they lead to
plotted objects that are indistinguishable. Typically, *variable*
transformation is done outside the graphics system and so the system
cannot provide transformation specific guides & decorations for the
plot. The :class:`trans` is aimed at being useful for *scale* and
*coordinate* transformations.
"""
import sys
import datetime
from collections.abc import Iterable
from types import MethodType

try:
    from zoneinfo import ZoneInfo
except ImportError:
    # python < 3.9
    from backports.zoneinfo import ZoneInfo

import numpy as np
import pandas as pd
import scipy.stats as stats
from matplotlib.dates import date2num, num2date

from .breaks import (extended_breaks, log_breaks, minor_breaks,
                     trans_minor_breaks, date_breaks,
                     timedelta_breaks)
from .formatters import mpl_format, date_format, timedelta_format
from .formatters import log_format


__all__ = ['asn_trans', 'atanh_trans', 'boxcox_trans',
           'modulus_trans',
           'datetime_trans', 'exp_trans', 'identity_trans',
           'log10_trans', 'log1p_trans', 'log2_trans',
           'log_trans', 'logit_trans', 'probability_trans',
           'probit_trans', 'reverse_trans', 'sqrt_trans',
           'timedelta_trans', 'pd_timedelta_trans',
           'pseudo_log_trans', 'reciprocal_trans',
           'trans', 'trans_new', 'gettrans']

UTC = ZoneInfo('UTC')


class trans:
    """
    Base class for all transforms

    This class is used to transform data and also tell the
    x and y axes how to create and label the tick locations.

    The key methods to override are :meth:`trans.transform`
    and :meth:`trans.inverse`. Alternately, you can quickly
    create a transform class using the :func:`trans_new`
    function.

    Parameters
    ----------
    kwargs : dict
        Attributes of the class to set/override

    Examples
    --------
    By default trans returns one minor break between every pair
    of major break

    >>> major = [0, 1, 2]
    >>> t = trans()
    >>> t.minor_breaks(major)
    array([0.5, 1.5])

    Create a trans that returns 4 minor breaks

    >>> t = trans(minor_breaks=minor_breaks(4))
    >>> t.minor_breaks(major)
    array([0.2, 0.4, 0.6, 0.8, 1.2, 1.4, 1.6, 1.8])
    """
    #: Aesthetic that the transform works on
    aesthetic = None

    #: Whether the untransformed data is numerical
    dataspace_is_numerical = True

    #: Limits of the transformed data
    domain = (-np.inf, np.inf)

    #: Callable to calculate breaks
    breaks_ = None

    #: Callable to calculate minor_breaks
    minor_breaks = None

    #: Function to format breaks
    format = staticmethod(mpl_format())

    def __init__(self, **kwargs):
        for attr in kwargs:
            if hasattr(self, attr):
                setattr(self, attr, kwargs[attr])
            else:
                raise KeyError(
                    "Unknown Parameter {!r}".format(attr))

        # Defaults
        if (self.breaks_ is None and
                'breaks_' not in kwargs):
            self.breaks_ = extended_breaks(n=5)

        if (self.minor_breaks is None and
                'minor_breaks' not in kwargs):
            self.minor_breaks = minor_breaks(1)

    @staticmethod
    def transform(x):
        """
        Transform of x
        """
        return x

    @staticmethod
    def inverse(x):
        """
        Inverse of x
        """
        return x

    def breaks(self, limits):
        """
        Calculate breaks in data space and return them
        in transformed space.

        Expects limits to be in *transform space*, this
        is the same space as that where the domain is
        specified.

        This method wraps around :meth:`breaks_` to ensure
        that the calculated breaks are within the domain
        the transform. This is helpful in cases where an
        aesthetic requests breaks with limits expanded for
        some padding, yet the expansion goes beyond the
        domain of the transform. e.g for a probability
        transform the breaks will be in the domain
        ``[0, 1]`` despite any outward limits.

        Parameters
        ----------
        limits : tuple
            The scale limits. Size 2.

        Returns
        -------
        out : array_like
            Major breaks
        """
        # clip the breaks to the domain,
        # e.g. probabilities will be in [0, 1] domain
        vmin = np.max([self.domain[0], limits[0]])
        vmax = np.min([self.domain[1], limits[1]])
        breaks = np.asarray(self.breaks_([vmin, vmax]))

        # Some methods(mpl_breaks, extended_breaks) that
        # calculate breaks take the limits as guide posts and
        # not hard limits.
        breaks = breaks.compress((breaks >= self.domain[0]) &
                                 (breaks <= self.domain[1]))
        return breaks


def trans_new(name, transform, inverse, breaks=None,
              minor_breaks=None, _format=None,
              domain=(-np.inf, np.inf), doc='', **kwargs):
    """
    Create a transformation class object

    Parameters
    ----------
    name : str
        Name of the transformation
    transform : callable ``f(x)``
        A function (preferably a `ufunc`) that computes
        the transformation.
    inverse : callable ``f(x)``
        A function (preferably a `ufunc`) that computes
        the inverse of the transformation.
    breaks : callable ``f(limits)``
        Function to compute the breaks for this transform.
        If None, then a default good enough for a linear
        domain is used.
    minor_breaks : callable ``f(major, limits)``
        Function to compute the minor breaks for this
        transform. If None, then a default good enough for
        a linear domain is used.
    _format : callable ``f(breaks)``
        Function to format the generated breaks.
    domain : array_like
        Domain over which the transformation is valid.
        It should be of length 2.
    doc : str
        Docstring for the class.
    **kwargs : dict
        Attributes of the transform, e.g if base is passed
        in kwargs, then `t.base` would be a valied attribute.

    Returns
    -------
    out : trans
        Transform class
    """
    def _get(func):
        if isinstance(func, (classmethod, staticmethod, MethodType)):
            return func
        else:
            return staticmethod(func)

    klass_name = '{}_trans'.format(name)

    d = {'transform': _get(transform),
         'inverse': _get(inverse),
         'domain': domain,
         '__doc__': doc,
         **kwargs}

    if breaks:
        d['breaks_'] = _get(breaks)

    if minor_breaks:
        d['minor_breaks'] = _get(minor_breaks)

    if _format:
        d['format'] = _get(_format)

    return type(klass_name, (trans,), d)


def log_trans(base=None, **kwargs):
    """
    Create a log transform class for *base*

    Parameters
    ----------
    base : float
        Base for the logarithm. If None, then
        the natural log is used.
    kwargs : dict
        Keyword arguments passed onto
        :func:`trans_new`. Should not include
        the `transform` or `inverse`.

    Returns
    -------
    out : type
        Log transform class
    """
    # transform function
    if base is None:
        name = 'log'
        base = np.exp(1)
        transform = np.log
    elif base == 10:
        name = 'log10'
        transform = np.log10
    elif base == 2:
        name = 'log2'
        transform = np.log2
    else:
        name = 'log{}'.format(base)

        def transform(x):
            return np.log(x)/np.log(base)

    # inverse function
    def inverse(x):
        return np.power(base, x)

    if 'domain' not in kwargs:
        kwargs['domain'] = (sys.float_info.min, np.inf)

    if 'breaks' not in kwargs:
        kwargs['breaks'] = log_breaks(base=base)

    kwargs['base'] = base
    kwargs['_format'] = log_format(base)

    _trans = trans_new(name, transform, inverse, **kwargs)

    if 'minor_breaks' not in kwargs:
        n = int(base) - 2
        _trans.minor_breaks = trans_minor_breaks(_trans, n=n)

    return _trans


log10_trans = log_trans(10, doc='Log 10 Transformation')
log2_trans = log_trans(2, doc='Log 2 Transformation')


def exp_trans(base=None, **kwargs):
    """
    Create a exponential transform class for *base*

    This is inverse of the log transform.

    Parameters
    ----------
    base : float
        Base of the logarithm
    kwargs : dict
        Keyword arguments passed onto
        :func:`trans_new`. Should not include
        the `transform` or `inverse`.

    Returns
    -------
    out : type
        Exponential transform class
    """
    # default to e
    if base is None:
        name = 'power_e'
        base = np.exp(1)
    else:
        name = 'power_{}'.format(base)

    # transform function
    def transform(x):
        return np.power(base, x)

    # inverse function
    def inverse(x):
        return np.log(x)/np.log(base)

    kwargs['base'] = base
    return trans_new(name, transform, inverse, **kwargs)


class log1p_trans(trans):
    """
    Log plus one Transformation
    """
    transform = staticmethod(np.log1p)
    inverse = staticmethod(np.expm1)


class identity_trans(trans):
    """
    Identity Transformation
    """
    pass


class reverse_trans(trans):
    """
    Reverse Transformation
    """
    transform = staticmethod(np.negative)
    inverse = staticmethod(np.negative)


class sqrt_trans(trans):
    """
    Square-root Transformation
    """
    transform = staticmethod(np.sqrt)
    inverse = staticmethod(np.square)
    domain = (0, np.inf)


class asn_trans(trans):
    """
    Arc-sin square-root Transformation
    """
    @staticmethod
    def transform(x):
        return 2*np.arcsin(np.sqrt(x))

    @staticmethod
    def inverse(x):
        return np.sin(x/2)**2


class atanh_trans(trans):
    """
    Arc-tangent Transformation
    """
    transform = staticmethod(np.arctanh)
    inverse = staticmethod(np.tanh)


def boxcox_trans(p, offset=0, **kwargs):
    r"""
    Boxcox Transformation

    The Box-Cox transformation is a flexible transformation, often
    used to transform data towards normality.

    The Box-Cox power transformation (type 1) requires strictly positive
    values and takes the following form for :math:`y \gt 0`:

    .. math::

        y^{(\lambda)} = \frac{y^\lambda - 1}{\lambda}

    When :math:`y = 0`, the natural log transform is used.

    Parameters
    ----------
    p : float
        Transformation exponent :math:`\lambda`.
    offset : int
        Constant offset. 0 for Box-Cox type 1, otherwise any
        non-negative constant (Box-Cox type 2).
        The default is 0. :func:`~mizani.transforms.modulus_trans`
        sets the default to 1.
    kwargs : dict
        Keyword arguments passed onto :func:`trans_new`. Should not
        include the `transform` or `inverse`.

    References
    ----------
    - Box, G. E., & Cox, D. R. (1964). An analysis of transformations.
      Journal of the Royal Statistical Society. Series B (Methodological),
      211-252. `<https://www.jstor.org/stable/2984418>`_
    - John, J. A., & Draper, N. R. (1980). An alternative family of
      transformations. Applied Statistics, 190-197.
      `<http://www.jstor.org/stable/2986305>`_

    See Also
    --------
    :func:`~mizani.transforms.modulus_trans`

    """
    def transform(x):
        x = np.asarray(x)
        if np.any((x + offset) < 0):
            raise ValueError(
                "boxcox_trans must be given only positive values. "
                "Consider using modulus_trans instead?"
            )
        if np.abs(p) < 1e-7:
            return np.log(x + offset)
        else:
            return ((x + offset)**p - 1)/p

    def inverse(x):
        x = np.asarray(x)
        if np.abs(p) < 1e-7:
            return np.exp(x) - offset
        else:
            return (x*p + 1) ** (1/p) - offset

    kwargs['p'] = p
    kwargs['offset'] = offset
    kwargs['name'] = kwargs.get('name', 'pow_{}'.format(p))
    kwargs['transform'] = transform
    kwargs['inverse'] = inverse
    return trans_new(**kwargs)


def modulus_trans(p, offset=1, **kwargs):
    r"""
    Modulus Transformation

    The modulus transformation generalises Box-Cox to work with
    both positive and negative values.

    When :math:`y \neq 0`

    .. math::

        y^{(\lambda)} = sign(y) * \frac{(|y| + 1)^\lambda - 1}{\lambda}

    and when :math:`y = 0`

    .. math::

        y^{(\lambda)} =  sign(y) * \ln{(|y| + 1)}

    Parameters
    ----------
    p : float
        Transformation exponent :math:`\lambda`.
    offset : int
        Constant offset. 0 for Box-Cox type 1, otherwise any
        non-negative constant (Box-Cox type 2).
        The default is 1. :func:`~mizani.transforms.boxcox_trans`
        sets the default to 0.
    kwargs : dict
        Keyword arguments passed onto :func:`trans_new`.
        Should not include the `transform` or `inverse`.

    References
    ----------
    - Box, G. E., & Cox, D. R. (1964). An analysis of transformations.
      Journal of the Royal Statistical Society. Series B (Methodological),
      211-252. `<https://www.jstor.org/stable/2984418>`_
    - John, J. A., & Draper, N. R. (1980). An alternative family of
      transformations. Applied Statistics, 190-197.
      `<http://www.jstor.org/stable/2986305>`_

    See Also
    --------
    :func:`~mizani.transforms.boxcox_trans`
    """
    if np.abs(p) < 1e-7:
        def transform(x):
            x = np.asarray(x)
            return np.sign(x) * np.log(np.abs(x) + offset)

        def inverse(x):
            x = np.asarray(x)
            return np.sign(x) * (np.exp(np.abs(x)) - offset)
    else:
        def transform(x):
            x = np.asarray(x)
            return np.sign(x) * ((np.abs(x) + offset)**p - 1) / p

        def inverse(x):
            x = np.asarray(x)
            return np.sign(x) * ((np.abs(x) * p + 1)**(1 / p) - offset)

    kwargs['p'] = p
    kwargs['offset'] = offset
    kwargs['name'] = kwargs.get('name', 'mt_pow_{}'.format(p))
    kwargs['transform'] = transform
    kwargs['inverse'] = inverse
    return trans_new(**kwargs)


def probability_trans(distribution, *args, **kwargs):
    """
    Probability Transformation

    Parameters
    ----------
    distribution : str
        Name of the distribution. Valid distributions are
        listed at :mod:`scipy.stats`. Any of the continuous
        or discrete distributions.
    args : tuple
        Arguments passed to the distribution functions.
    kwargs : dict
        Keyword arguments passed to the distribution functions.

    Notes
    -----
    Make sure that the distribution is a good enough
    approximation for the data. When this is not the case,
    computations may run into errors. Absence of any errors
    does not imply that the distribution fits the data.
    """
    cdists = {k for k in dir(stats)
              if hasattr(getattr(stats, k), 'cdf')}
    if distribution not in cdists:
        msg = "Unknown distribution '{}'"
        raise ValueError(msg.format(distribution))

    try:
        doc = kwargs.pop('_doc')
    except KeyError:
        doc = ''

    try:
        name = kwargs.pop('_name')
    except KeyError:
        name = 'prob_{}'.format(distribution)

    def transform(x):
        return getattr(stats, distribution).cdf(x, *args, **kwargs)

    def inverse(x):
        return getattr(stats, distribution).ppf(x, *args, **kwargs)

    return trans_new(name,
                     transform, inverse, domain=(0, 1),
                     doc=doc)


logit_trans = probability_trans('logistic', _name='logit',
                                _doc='Logit Transformation')
probit_trans = probability_trans('norm', _name='norm',
                                 _doc='Probit Transformation')


class datetime_trans(trans):
    """
    Datetime Transformation

    Parameters
    ----------
    tz : str | ZoneInfo
        Timezone information

    Examples
    --------
    >>> # from zoneinfo import ZoneInfo
    >>> # from backports.zoneinfo import ZoneInfo  # for python < 3.9
    >>> UTC = ZoneInfo("UTC")
    >>> EST = ZoneInfo("EST")
    >>> t = datetime_trans(EST)
    >>> x = datetime.datetime(2022, 1, 20, tzinfo=UTC)
    >>> x2 = t.inverse(t.transform(x))
    >>> x == x2
    True
    >>> x.tzinfo == x2.tzinfo
    False
    >>> x.tzinfo.key
    'UTC'
    >>> x2.tzinfo.key
    'EST'
    """
    dataspace_is_numerical = False
    domain = (
        datetime.datetime(datetime.MINYEAR, 1, 1, tzinfo=UTC),
        datetime.datetime(datetime.MAXYEAR, 12, 31, tzinfo=UTC)
    )
    breaks_ = staticmethod(date_breaks())
    format = staticmethod(date_format())
    tz = None

    def __init__(self, tz=None, **kwargs):
        if isinstance(tz, str):
            tz = ZoneInfo(tz)

        super().__init__(**kwargs)
        self.tz = tz

    def transform(self, x):
        """
        Transform from date to a numerical format
        """
        if isinstance(x, Iterable) and not len(x):
            return []

        try:
            tz = x[0].tzinfo
        except TypeError:
            tz = x.tzinfo
        except AttributeError:
            tz = None

        if tz and self.tz is None:
            self.tz = tz

        return date2num(x)

    def inverse(self, x):
        """
        Transform to date from numerical format
        """
        return num2date(x, tz=self.tz)

    @property
    def tzinfo(self):
        """
        Alias of `tz`
        """
        return self.tz


class timedelta_trans(trans):
    """
    Timedelta Transformation
    """
    dataspace_is_numerical = False
    domain = (datetime.timedelta.min, datetime.timedelta.max)
    breaks_ = staticmethod(timedelta_breaks())
    format = staticmethod(timedelta_format())

    @staticmethod
    def transform(x):
        """
        Transform from Timeddelta to numerical format
        """
        # microseconds
        try:
            x = np.array([_x.total_seconds()*10**6 for _x in x])
        except TypeError:
            x = x.total_seconds()*10**6
        return x

    @staticmethod
    def inverse(x):
        """
        Transform to Timedelta from numerical format
        """
        try:
            x = [datetime.timedelta(microseconds=i) for i in x]
        except TypeError:
            x = datetime.timedelta(microseconds=x)
        return x


class pd_timedelta_trans(trans):
    """
    Pandas timedelta Transformation
    """
    dataspace_is_numerical = False
    domain = (pd.Timedelta.min, pd.Timedelta.max)
    breaks_ = staticmethod(timedelta_breaks())
    format = staticmethod(timedelta_format())

    @staticmethod
    def transform(x):
        """
        Transform from Timeddelta to numerical format
        """
        # nanoseconds
        try:
            x = np.array([_x.value for _x in x])
        except TypeError:
            x = x.value
        return x

    @staticmethod
    def inverse(x):
        """
        Transform to Timedelta from numerical format
        """
        try:
            x = [pd.Timedelta(int(i)) for i in x]
        except TypeError:
            x = pd.Timedelta(int(x))
        return x


class reciprocal_trans(trans):
    """
    Reciprocal Transformation
    """

    @staticmethod
    def transform(x):
        x = np.asarray(x)
        return 1 / x

    @staticmethod
    def inverse(x):
        x = np.asarray(x)
        return 1 / x


def pseudo_log_trans(sigma=1, base=None, **kwargs):
    """
    Pseudo-log transformation

    A transformation mapping numbers to a signed logarithmic
    scale with a smooth transition to linear scale around 0.

    Parameters
    ----------
    sigma : float
        Scaling factor for the linear part.
    base : int
        Approximate logarithm used. If None, then
        the natural log is used.
    kwargs : dict
        Keyword arguments passed onto
        :func:`trans_new`. Should not include
        the `transform` or `inverse`.
    """
    if base is None:
        base = np.exp(1)

    def transform(x):
        x = np.asarray(x)
        return np.arcsinh(x/(2*sigma)) / np.log(base)

    def inverse(x):
        x = np.asarray(x)
        return 2 * sigma * np.sinh(x * np.log(base))

    kwargs['base'] = base
    kwargs['sigma'] = sigma
    _trans = trans_new('pseudo_log', transform, inverse, **kwargs)

    if 'minor_breaks' not in kwargs:
        n = int(base) - 2
        _trans.minor_breaks = trans_minor_breaks(_trans, n=n)

    return _trans


def gettrans(t):
    """
    Return a trans object

    Parameters
    ----------
    t : str | callable | type | trans
        name of transformation function

    Returns
    -------
    out : trans
    """
    obj = t
    # Make sure trans object is instantiated
    if isinstance(obj, str):
        name = '{}_trans'.format(obj)
        obj = globals()[name]()
    if callable(obj):
        obj = obj()
    if isinstance(obj, type):
        obj = obj()

    if not isinstance(obj, trans):
        raise ValueError("Could not get transform object.")

    return obj



from itertools import chain
from collections import OrderedDict, defaultdict
from collections.abc import Iterator

import numpy as np

__all__ = ['round_any', 'min_max', 'match',
           'precision', 'first_element', 'multitype_sort',
           'same_log10_order_of_magnitude',
           'identity', 'get_categories'
           ]

DISCRETE_KINDS = 'ObUS'
CONTINUOUS_KINDS = 'ifuc'

SECONDS = OrderedDict([
    ('ns', 1e-9),        # nanosecond
    ('us', 1e-6),        # microsecond
    ('ms', 1e-3),        # millisecond
    ('s', 1),            # second
    ('m', 60),           # month
    ('h', 3600),         # hour
    ('d', 24*3600),      # day
    ('w', 7*24*3600),    # week
    ('M', 31*24*3600),   # month
    ('y', 365*24*3600),  # year
])

NANOSECONDS = OrderedDict([
    ('ns', 1),             # nanosecond
    ('us', 1e3),           # microsecond
    ('ms', 1e6),           # millisecond
    ('s', 1e9),            # second
    ('m', 60e9),           # month
    ('h', 3600e9),         # hour
    ('d', 24*3600e9),      # day
    ('w', 7*24*3600e9),    # week
    ('M', 31*24*3600e9),   # month
    ('y', 365*24*3600e9),  # year
])


def round_any(x, accuracy, f=np.round):
    """
    Round to multiple of any number.
    """
    if not hasattr(x, 'dtype'):
        x = np.asarray(x)

    return f(x / accuracy) * accuracy


def min_max(x, na_rm=False, finite=True):
    """
    Return the minimum and maximum of x

    Parameters
    ----------
    x : array_like
        Sequence
    na_rm : bool
        Whether to remove ``nan`` values.
    finite : bool
        Whether to consider only finite values.

    Returns
    -------
    out : tuple
        (minimum, maximum) of x
    """
    if not hasattr(x, 'dtype'):
        x = np.asarray(x)

    if na_rm and finite:
        x = x[np.isfinite(x)]
    elif not na_rm and np.any(np.isnan(x)):
        return np.nan, np.nan
    elif na_rm:
        x = x[~np.isnan(x)]
    elif finite:
        x = x[~np.isinf(x)]

    if (len(x)):
        return np.min(x), np.max(x)
    else:
        return float('-inf'), float('inf')


def match(v1, v2, nomatch=-1, incomparables=None, start=0):
    """
    Return a vector of the positions of (first)
    matches of its first argument in its second.

    Parameters
    ----------
    v1: array_like
        Values to be matched

    v2: array_like
        Values to be matched against

    nomatch: int
        Value to be returned in the case when
        no match is found.

    incomparables: array_like
        Values that cannot be matched. Any value in ``v1``
        matching a value in this list is assigned the nomatch
        value.
    start: int
        Type of indexing to use. Most likely 0 or 1
    """
    v2_indices = {}
    for i, x in enumerate(v2):
        if x not in v2_indices:
            v2_indices[x] = i

    v1_to_v2_map = [nomatch] * len(v1)
    skip = set(incomparables) if incomparables else set()
    for i, x in enumerate(v1):
        if x in skip:
            continue

        try:
            v1_to_v2_map[i] = v2_indices[x] + start
        except KeyError:
            pass

    return v1_to_v2_map


def precision(x):
    """
    Return the precision of x

    Parameters
    ----------
    x : array_like | numeric
        Value(s) whose for which to compute the precision.

    Returns
    -------
    out : numeric
        The precision of ``x`` or that the values in ``x``.

    Notes
    -----
    The precision is computed in base 10.

    Examples
    --------
    >>> precision(0.08)
    0.01
    >>> precision(9)
    1
    >>> precision(16)
    10
    """
    from .bounds import zero_range

    rng = min_max(x, na_rm=True)
    if zero_range(rng):
        span = np.abs(rng[0])
    else:
        span = np.diff(rng)[0]

    if span == 0:
        return 1
    else:
        return 10 ** int(np.floor(np.log10(span)))


def first_element(obj):
    """
    Return the first element of `obj`

    Parameters
    ----------
    obj : iterable
        Should not be an iterator

    Returns
    -------
    out : object
        First element of `obj`. Raise a class:`StopIteration`
        exception if `obj` is empty.
    """
    if isinstance(obj, Iterator):
        raise RuntimeError(
            "Cannot get the first element of an iterator")
    return next(iter(obj))


def multitype_sort(a):
    """
    Sort elements of multiple types

    x is assumed to contain elements of different types, such that
    plain sort would raise a `TypeError`.

    Parameters
    ----------
    a : array-like
        Array of items to be sorted

    Returns
    -------
    out : list
        Items sorted within their type groups.
    """
    types = defaultdict(list)
    numbers = {int, float, complex}

    for x in a:
        t = type(x)
        if t in numbers:
            types['number'].append(x)
        else:
            types[t].append(x)

    for t in types:
        types[t] = np.sort(types[t])

    return list(chain.from_iterable(types[t] for t in types))


def same_log10_order_of_magnitude(x, delta=0.1):
    """
    Return true if range is approximately in same order of magnitude

    For example these sequences are in the same order of magnitude:

        - [1, 8, 5]     # [1, 10)
        - [35, 20, 80]  # [10 100)
        - [232, 730]    # [100, 1000)

    Parameters
    ----------
    x : array-like
         Values in base 10. Must be size 2 and
        ``rng[0] <= rng[1]``.
    delta : float
        Fuzz factor for approximation. It is multiplicative.
    """
    dmin = np.log10(np.min(x)*(1-delta))
    dmax = np.log10(np.max(x)*(1+delta))
    return np.floor(dmin) == np.floor(dmax)


def identity(*args):
    """
    Return whatever is passed in
    """
    return args if len(args) > 1 else args[0]


def get_categories(x):
    """
    Return the categories of x

    Parameters
    ----------
    x : category_like
        Input Values

    Returns
    -------
    out : Index
        Categories of x
    """
    try:
        return x.cat.categories  # series
    except AttributeError:
        try:
            return x.categories   # plain categorical
        except AttributeError:
            raise TypeError("x is the wrong type, it has no categories")



import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from mizani.bounds import rescale
from mizani.scale import scale_continuous, scale_discrete
from mizani.transforms import identity_trans


def test_scale_continuous():
    x = np.arange(11)
    # apply
    scaled = scale_continuous.apply(x, rescale)
    npt.assert_allclose(scaled, x*0.1)
    npt.assert_allclose(scaled, x*0.1)

    # Train
    limits = scale_continuous.train(x)
    npt.assert_allclose(limits, [0, 10])
    # Additional training
    limits = scale_continuous.train(np.arange(-4, 11), limits)
    npt.assert_allclose(limits, [-4, 10])
    limits = scale_continuous.train([], limits)
    npt.assert_allclose(limits, [-4, 10])

    # branches #
    scaled = scale_continuous.apply(x, rescale,
                                    trans=identity_trans())
    mapped = scale_continuous.map(x, list, [0, 10])
    npt.assert_allclose(mapped, x*0.1)

    with pytest.raises(TypeError):
        # discrete data
        limits = scale_continuous.train(['a', 'b', 'c'])


def test_scale_discrete():
    def assert_equal_with_nan(lst1, lst2):
        assert lst1[:-1] == lst2[:-1] and np.isnan(lst2[-1])

    def SCategorical(*args, **kwargs):
        return pd.Series(pd.Categorical(*args, **kwargs))

    x = ['a', 'b', 'c', 'a']
    # apply
    scaled = scale_discrete.apply(x, np.arange)
    npt.assert_allclose(scaled, [0, 1, 2, 0])

    # Train
    limits = scale_discrete.train(x)
    assert limits == ['a', 'b', 'c']

    # Additional training
    limits = scale_discrete.train(['b', 'c', 'd'], limits)
    assert limits == ['a', 'b', 'c', 'd']
    limits = scale_discrete.train([], limits)
    assert limits == ['a', 'b', 'c', 'd']
    limits = scale_discrete.train([None, 'e'], limits)
    assert_equal_with_nan(limits, ['a', 'b', 'c', 'd', 'e', np.nan])

    # Train (Deal with missing values)
    x1 = ['a', 'b', np.nan, 'c']
    x2 = pd.Categorical([1, 2, 3, np.nan])
    limits = scale_discrete.train(x1, na_rm=True)
    assert limits == ['a', 'b', 'c']

    limits = scale_discrete.train(x1, na_rm=False)
    assert_equal_with_nan(limits, ['a', 'b', 'c', np.nan])

    limits = scale_discrete.train(x2, na_rm=True)
    assert limits == [1, 2, 3]

    limits = scale_discrete.train(x2, na_rm=False)
    assert_equal_with_nan(limits, [1, 2, 3, np.nan])

    # Double trainning with nan does not add another nan
    limits2 = scale_discrete.train(x2, limits, na_rm=False)
    assert_equal_with_nan(limits2, [1, 2, 3, np.nan])

    # branches #

    with pytest.raises(TypeError):
        # continuous data
        limits = scale_discrete.train([1, 2, 3])

    x = pd.Categorical(['a', 'b'])
    limits = scale_discrete.train(x)
    assert limits == ['a', 'b']

    x = pd.Series(['a', 'b', 'c'], dtype='category')
    limits = scale_discrete.train(x[:1], drop=True)
    assert limits == ['a']

    limits = scale_discrete.train(x[:2], drop=False)
    assert limits == ['a', 'b', 'c']

    # Disrete Scale training maintains order of categoricals
    cats = ['x0', 'x1', 'x2', 'x3', 'x4']
    s1 = SCategorical(['x1', 'x2'], categories=cats)
    s2 = SCategorical(['x0', 'x2'], categories=cats)
    limits = scale_discrete.train(s1, drop=True)
    limits = scale_discrete.train(s2, limits, drop=True)
    assert limits == ['x0', 'x1', 'x2']

    # Trainning on mixed categories, the last data determines
    # the location of a value that is in two categoricals
    # eg. a & e are ordered right!
    x1 = pd.Categorical(['a', 'b', 'c', 'e'])
    x2 = pd.Categorical(['d', 'f', 'e', 'a'])
    limits = scale_discrete.train(x1)
    limits = scale_discrete.train(x2, old=limits)
    # assert limits == list('abcedf')
    assert limits == list('adefbc')



import pytest
import numpy as np
import numpy.testing as npt
import matplotlib as mpl

from mizani.palettes import (hls_palette, husl_palette, rescale_pal,
                             area_pal, abs_area, grey_pal, hue_pal,
                             brewer_pal, gradient_n_pal, cmap_pal,
                             cmap_d_pal,
                             desaturate_pal, manual_pal, xkcd_palette,
                             crayon_palette, cubehelix_pal, identity_pal)
from mizani.palettes import ratios_to_colors


def test_hls_palette():
    colors = hls_palette(10)
    assert len(colors) == 10
    assert all(len(c) == 3 for c in colors)


def test_husl_palette():
    colors = husl_palette(5)
    assert len(colors) == 5
    assert all(len(c) == 3 for c in colors)


def test_rescale_pal():
    palette = rescale_pal()
    x = np.arange(0, 1+.01, 0.1)
    result = palette(x)
    assert min(result) == 0.1
    assert max(result) == 1

    palette = rescale_pal((20, 100))
    result = palette(x)
    assert min(result) == 20
    assert max(result) == 100


def test_area_pal():
    palette = area_pal((0, 10))
    x = np.arange(0, 11)
    xsq = (x * .1)**2
    result = palette(xsq)
    npt.assert_allclose(result, x)


def test_abs_area():
    x = np.arange(0, 1.03, .1)**2
    palette = abs_area(5)
    result = palette(x)
    assert min(result) == 0
    assert max(result) == 5


def test_grey_pal():
    palette = grey_pal()
    result = palette(5)
    # Same rgb values
    assert all(s[1:3]*3 == s[1:] for s in result)


def test_hue_pal():
    palette = hue_pal()
    result = palette(5)
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    # branches #
    with pytest.raises(ValueError):
        hue_pal(.1, 2.3, 3)

    with pytest.raises(ValueError):
        hue_pal(color_space='slh')


def test_brewer_pal():
    result = brewer_pal()(5)
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    result = brewer_pal('qual', 2)(5)
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    result = brewer_pal('div', 2)(5)
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    with pytest.raises(ValueError):
        brewer_pal('div', 200)(5)

    result = brewer_pal('seq', 'Greens')(5)
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    with pytest.warns(UserWarning):
        brewer_pal()(100)

    result = brewer_pal('seq', 'Blues')(2)
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    result1 = brewer_pal('seq', 'Blues', direction=1)(2)
    result2 = brewer_pal('seq', 'Blues', direction=-1)(2)
    assert result1 == result2[::-1]

    with pytest.raises(ValueError):
        brewer_pal(direction=-2)(100)


def test_gradient_n_pal():
    palette = gradient_n_pal(['red', 'blue'])
    result = palette([0, .25, .5, .75, 1])
    assert result[0].lower() == '#ff0000'
    assert result[-1].lower() == '#0000ff'
    assert palette([0])[0].lower() == '#ff0000'

    # symmetric gradient
    palette = gradient_n_pal(['red', 'blue', 'red'], [0, 0.5, 1])
    result = palette([0.2, 0.8])
    assert result[0] == result[1]

    # single color
    result = palette(.25)
    assert result.lower().startswith('#')


def test_cmap_pal():
    palette = cmap_pal('viridis')
    result = palette([0, .25, .5, .75, 1])
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    with pytest.warns(FutureWarning):
        cmap_pal('viridis', 5)


def test_cmap_d_pal():
    palette = cmap_d_pal('viridis')
    result = palette(6)
    assert all(s[0] == '#' and len(s) == 7 for s in result)
    assert len(result) == 6

    # More colors than palette
    palette = cmap_d_pal('Accent')
    result = palette(5)
    assert all(s[0] == '#' and len(s) == 7 for s in result)

    with pytest.raises(ValueError):
        result = palette(10)

    # Bad palette
    with pytest.raises(ValueError):
        palette = cmap_d_pal('Greens')

    with pytest.warns(FutureWarning):
        cmap_d_pal('viridis', 5)


def test_desaturate_pal():
    x = [0, .25, .5, .75, 1]
    # When desaturating pure red, green and blue
    # the other 2 colors get matching values,
    # we test that.
    result = desaturate_pal('red', .1)(x)
    assert all(s[3:5] == s[5:] for s in result)

    result = desaturate_pal('green', .2)(x)
    assert all(s[1:3] == s[-2:] for s in result)

    result = desaturate_pal('blue', .3)(x)
    assert all(s[1:3] == s[3:5] for s in result)

    result = desaturate_pal('blue', .3, reverse=True)(x)
    assert all(s[1:3] == s[3:5] for s in result)

    with pytest.raises(ValueError):
        desaturate_pal('green', 2.3)


def test_manual_pal():
    size = 5
    n = 3
    values = list(range(size))
    palette = manual_pal(values)
    result = palette(n)
    assert result == values[:n]

    with pytest.warns(UserWarning):
        result = palette(size+4)


def test_xkcd_palette():
    values = xkcd_palette(['apple green', 'red', 'tan brown'])
    assert len(values) == 3
    assert all(s[0] == '#' and len(s) == 7 for s in values)


def test_crayon_palette():
    values = crayon_palette(['banana mania', 'red', 'yellow'])
    assert len(values) == 3
    assert all(s[0] == '#' and len(s) == 7 for s in values)


def test_cubehelix_pal():
    palette = cubehelix_pal()
    values = palette(5)
    assert len(values) == 5
    assert all(s[0] == '#' and len(s) == 7 for s in values)


def test_identity_pal():
    palette = identity_pal()

    x = [1, 2, 3]
    values = palette(x)
    assert values == x

    value = palette(10)
    assert value == 10


def test_ratios_to_colors():
    x = 0.5
    result = ratios_to_colors(x, mpl.colormaps['viridis'])
    assert result[0] == '#'
